#!/usr/bin/env python
"""Contains functions for UROPA configuration."""
import json
import ast
import os
import subprocess
from textwrap import dedent
import numpy as np


def howtoconfig():
    """Defines the epilog that is given when help is requested."""
    epilog = dedent("""\
    UROPA is a peak annotation tool facilitating the analysis of next-generation sequencing methods for
    chromatin biology, like ChIPseq or ATACseq. There are already different peak annotation tools, like
    HOMER or ChIPpeakAnno, but the advantage of UROPA is, that it can easily be fitted to your requirements.
    UROPA was developed as an open source analysis pipeline for peaks generated from any peak caller.

    Please cite upon usage:
    Kondili M, Fust A, Preussner J, Kuenne C, Braun T and Looso M. UROPA: A tool for Universal RObust Peak Annotation.
    Scientific Reports 7 (2017), doi: 10.1038/s41598-017-02464-y

    All parameters and paths to input or output files should be reported in a JSON configuration file.
    The configuration file should at least contain paths for bed and GTF files:

    {
    "queries": [],
    "bed": "/path/to/bed/file.bed",
    "gtf": "/path/to/annotation/file.gtf"
    }

    Different query types can be defined using the queries key:

    {
    "queries": [
      {...},
      {...}],
    "bed": "/path/to/bed/file.bed",
    "gtf": "/path/to/annotation/file.gtf"
    }

    Optionally, the priority key can be used to fine tune UROPAs behaviour:

    {
    "queries": [
      {...},
      {...}],
    "bed": "/path/to/bed/file.bed",
    "gtf": "/path/to/annotation/file.gtf",
    "priority": "True"
    }

    Please visit http://uropa-manual.readthedocs.io/config.html for detailed information on configuration.
    """)
    return epilog


def parse_json(infile):
    """Read a json file"""
    assert isinstance(infile, str), 'Argument {0} of wrong type ({1}), should be {2}!'.format(
        'column', type(infile), 'str')

    with open(infile, 'r', encoding='utf-8') as f:
        return ast.literal_eval(json.dumps(json.load(f)))


def column_from_file(file, column, log=None):
    """Extracts a given column from a file, and returns unified values as list."""
    assert isinstance(file, str), 'Argument {0} of wrong type ({1}), should be {2}!'.format(
        'column', type(file), 'str')
    assert isinstance(column, int), 'Argument {0} of wrong type ({1}), should be {2}!'.format(
        'column', type(column), 'int')

    cmd = 'cut -f' + str(column) + ' ' + str(file) + \
        ' | sort | uniq | grep -v "^#"'

    try:
        vals = str(subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True), encoding='utf-8')
    except subprocess.CalledProcessError:
        if log is not None:
            log.warning("File {} might be empty or has not enough columns".format(file))
        return([])
    return([v for v in vals.split('\n') if v != ""])


def parse_parameters(config, log=None):
    """Fills the configuration with default values. Writes a warning to logs, if unknown keys are detected."""
    defaults = {"priority": "False", "bed": "no_peaks.bed",
                "gtf": "no_annotation.gtf"}  # , "bigwig": "none.bw"
    keys = list(defaults.keys())
    values = [config[x] if x in config else defaults[x] for x in keys]

    if log is not None:
        unknown = [k for k in list(config.keys()) if k not in keys and k !=
                   "queries"]
        if any(unknown):
            log.warning(
                "Unknown keys detected in configuration: {}".format(unknown))
    parameters = dict(list(zip(keys, values)))
    return parameters


def parse_queries(config, gtf_feat, log=None):
    """Adds in defaults where query parameters are missing."""
    defaults = {"feature": gtf_feat, "strand": "ignore", "show.attributes": "None", "filter.attribute": "None",
                "attribute.value": "None", "distance": 100000, "feature.anchor": ["start", "center", "end"],
                "direction": "any_direction", "internals": "False", "priority": "False"}
    keys = list(defaults.keys())

    try:
        query_list = config["queries"]
    except KeyError:
        query_list = [defaults]
        if not log is None:
            log.warning("No 'queries' key given, so the default values will be used: {}".format(query_list))

    if len(query_list) == 0:
        if not log is None:
            log.info('Empty queries given ("queries":[]). Will use defaults.')
        query_list = [defaults]

    if not isinstance(query_list, list):
        query_list = [query_list]

    def give_val(q):
        return [q[x] if x in q and q[x] != "" else defaults[x] for x in keys]

    def make_list(l):
        return [[v] if not isinstance(v, list) else v for v in l]

    vals = [give_val(l) for l in query_list]
    values = [make_list(l) for l in vals]

    if not log is None:
        unknown = [k for k in list(config.keys()) if k not in list(set(['gtf', 'bed', 'queries']).union(keys))]
        if any(unknown):
            log.warning("Unknown keys detected in configuration: {}".format(unknown))

    queries = [dict(list(zip(keys, x))) for x in values]
    return queries


def remove_invalid_queries(queries, log=None):
    """Removes queries that have multiple attribute.filter or filter.value values and more than two distance constraints."""
    # Validate distance constraints
    has_valid_distance = [len(q["distance"]) < 3 for q in queries]
    if not all(has_valid_distance) and log is not None:
        log.warning("Queries with invalid distances present! Affected queries: {}".format(
            [i for i, x in enumerate(has_valid_distance) if not x]))

    # Validate query attributes
    has_valid_attributes = [False if (q["filter.attribute"] != ['None'] and q["attribute.value"] == [
        'None']) or (q["filter.attribute"] == ['None'] and q["attribute.value"] != ['None']) else True for q in queries]

    if not all(has_valid_attributes) and log is not None:
        log.warning("Queries with invalid filter.attribute and attribute.value pairings present! Affected queries: {}".format(
            [i for i, x in enumerate(has_valid_attributes) if not x]))

    # Validate strand attribute
    has_valid_strand = [True if q["strand"] in [["ignore"], ["same"], ["opposite"]] else False for q in queries]

    if not all(has_valid_strand) and log is not None:
        log.warning("Queries with invalid strand values present! Affected queries: {}".format(
            [i for i, x in enumerate(has_valid_strand) if not x]))

    # Validate query attribute lengths
    has_valid_attribute_lengths = [False if len(
        q["filter.attribute"]) > 1 or len(q["attribute.value"]) > 1 else True for q in queries]
    if not all(has_valid_attribute_lengths) and log is not None:
        log.warning("Queries with more than one value for either filter.attribute or attribute.value present! Affected queries: {}".format(
            [i for i, x in enumerate(has_valid_attribute_lengths) if not x]))

    invalid_queries = list(np.unique([i for i, x in enumerate(has_valid_attributes) if not x] + [i for i, x in enumerate(
        has_valid_attribute_lengths) if not x] + [i for i, x in enumerate(has_valid_distance) if not x] + [i for i, x in enumerate(has_valid_strand) if not x]))
    return([queries[i] for i in range(0, len(queries)) if not i in invalid_queries])


def parse_first_gtf_line(gtf):
    """Removes comment lines, reads first line to check for prefix chr. Returns if chr was found and the number of columns."""
    f = open(gtf, "r", encoding='utf-8')
    is_comment = True
    while is_comment:
        line = f.readline()
        is_comment = line.startswith("#")

    has_chr = line.startswith('chr')
    num_cols = len(line.split("\t"))
    return(has_chr, num_cols)


def cut_gtf_perFeat(gtf, features, prefix):
    """Removes lines with features not in features from a gtf file."""
    gtf_per_feat = prefix + os.path.basename(gtf).split(".gtf")[0] + "_cut_per_feat.gtf"
    feat2cut = np.unique(features)

    f = open(gtf, "r", encoding='utf-8')
    is_comment = True
    while is_comment:
        line = f.readline()
        is_comment = line.startswith("#")

    lines = f.readlines()
    gtf_query_feat = [li for li in lines if li.split("\t")[2] in feat2cut]

    # List of lines containing only selected features from query
    with open(gtf_per_feat, "w") as cutgtf:
        cutgtf.write("".join(gtf_query_feat))
    return gtf_per_feat



"""
uropa.py: UROPA - Universal RObust Peak Annotator

@authors: Maria Kondili, Jens Preussner and Annika Fust
@license: MIT
@version: 2.0.0-alpha
@maintainer: Mario Looso
@email: mario.looso@mpi-bn.mpg.de
"""


import os
import sys
import json
import glob
import argparse
import logging
import datetime
import shutil
import subprocess as sp
import multiprocessing as mp
from itertools import chain
from functools import partial
from functools import reduce

import numpy as np

from . import config as cfg
from . import overlaps as ovls
from . import annotation as ant

#if __name__ == "__main__":
def main():
	parser = argparse.ArgumentParser(
		prog="uropa",
		description='UROPA - Universal RObust Peak Annotator.',
		epilog=cfg.howtoconfig(),
		formatter_class=argparse.RawDescriptionHelpFormatter)
	parser.add_argument(
		"-i",
		"--input",
		help="filename of configuration file",
		required=True,
		action="store",
		metavar="config.json")
	parser.add_argument(
		"-p",
		"--prefix",
		dest="prefix",
		help="prefix for result file names (defaults to basename of input file)",
		required=False,
		action="store",
		metavar="prefix")
	parser.add_argument(
		"-r",
		"--reformat",
		help="create an additional compact and line-reduced table as result file",
		required=False,
		action="store_true")
	parser.add_argument(
		"-s",
		"--summary",
		help="filename of additional visualisation of results in graphical format",
		required=False,
		action="store_true")
	parser.add_argument(
		"-t",
		"--threads",
		help="multiprocessed run: n = number of threads to run annotation process",
		type=int,
		required=False,
		action="store",
		metavar="n",
		default=1)
	parser.add_argument(
		"--add-comments",
		help="add comment lines to output files",
		required=False,
		action="store_true")
	parser.add_argument(
		"-l",
		"--log",
		help="log file name for messages and warnings",
		action="store",
		metavar="uropa.log")
	parser.add_argument(
		"-d",
		"--debug",
		help="print verbose messages (for debugging) to stdout and log",
		required=False,
		action="store_true")
	parser.add_argument(
		"-v",
		"--version",
		help="prints the version and exits",
		action="version",
		version="%(prog)s 1.2.1")
	args = parser.parse_args()

	config = args.input

	# Configure logging
	logger = logging.getLogger(__name__)
	logger.setLevel(logging.DEBUG)

	streamHandle = logging.StreamHandler()
	loggerFormat = logging.Formatter('[%(levelname)s] - %(message)s')
	streamHandle.setFormatter(loggerFormat)

	if args.debug:
		streamHandle.setLevel(logging.DEBUG)
	else:
		streamHandle.setLevel(logging.WARNING)
	logger.addHandler(streamHandle)

	if args.log is not None:
		logpath = os.path.dirname(args.log)
		if not os.path.exists(logpath) and logpath != '':
			try:
				os.makedirs(logpath)
			except OSError:
				logger.error("Could not create directory for log file {}".format(logpath))
		try:
			fileHandle = logging.FileHandler(args.log, mode='w')
			fileHandle.setLevel(logging.DEBUG)
			fileHandle.setFormatter(loggerFormat)
			logger.addHandler(fileHandle)
		except IOError:
			logger.error("Could not create log file {}".format(args.log))

	if args.prefix is not None:
		prefixpath = os.path.dirname(args.prefix)
		if not os.path.exists(prefixpath):
			if not prefixpath == '':
				try:
					os.makedirs(prefixpath)
				except IOError:
					logger.error("Could not create directory {} for output".format(prefixpath))
			else:
				prefixpath = '.'
		outdir = prefixpath + '/' + os.path.basename(args.prefix) + '_'
	else:
		outdir = "./" + os.path.splitext(os.path.basename(args.input))[0] + '_'

	logger.debug("Directory for output files is {}".format(outdir))
	logger.info("Start time: %s", datetime.datetime.now().strftime("%d.%m.%Y %H:%M"))

	try:
		cfg_dict = cfg.parse_json(config)
	except IOError:
		logger.error("File %s does not exists or is not readable.", config)
		sys.exit()
	except ValueError as e:
		logger.error("File %s contains malformed JSON. %s", config, e)
		sys.exit()

	parameters = cfg.parse_parameters(cfg_dict, logger)
	priority = parameters["priority"]
	annot_gtf = parameters["gtf"]
	peaks_bed = parameters["bed"]

	try:
		gtf_has_chr, gtf_cols = cfg.parse_first_gtf_line(annot_gtf)
	except IOError:
		logger.error("File %s does not exist or is not readable.", annot_gtf)
		sys.exit()
	
	# > Check GTF format
	if gtf_cols == 9:
		logger.info("GTF file format validated. Extracting all features from GTF now...")
	else:
		logger.warning("File %s is not a proper GTF file!", annot_gtf)

	gtf_feat = cfg.column_from_file(annot_gtf, 3, logger)
	
	print("gtf_feat = ", gtf_feat)

	if len(gtf_feat) < 1:
		logger.error("No features found in file {} for annotation.".format(annot_gtf))
		sys.exit()

	if not os.path.exists(peaks_bed):
		logger.error("File %s does not exists or is not readable.", peaks_bed)
		sys.exit()

	# Sanity check for other parameters
	accepted_val = ['T', 'True', 'TRUE', 'Yes', 'YES', 'Y', 'yes', 'F', 'False', 'FALSE', 'No', 'NO', 'N', 'no']
	if priority not in accepted_val:
		logger.error("Priority value should be one of the values from %s", accepted_val)
		sys.exit()

	# All parameters of Queries, filled-in with default values where necessary.
	all_queries = cfg.parse_queries(cfg_dict, gtf_feat, logger)
	logger.info("Number of all queries for parametric peak annotation: %s", len(all_queries))
	queries = cfg.remove_invalid_queries(all_queries, logger)
	logger.info("Number of valid queries for parametric peak annotation: %s", len(queries))

	# Validate query features
	query_features = list(chain.from_iterable([q["feature"] for q in queries]))
	feat_not_valid = [f for f in query_features if f not in gtf_feat]
	feat_valid = [f for f in query_features if f not in feat_not_valid]

	if feat_not_valid:
		logger.warning("Invalid features present (found in a query but not in the GTF): %s", ','.join(feat_not_valid))

	# > Cut gtf according to features of config, and index the cut.gtf for Faster Annotation
	# ! Attention : In the UCSC transformed gtf files,the feature is put by default "tfbs"
	if len(gtf_feat) > 1:
		gtf_cut_file = cfg.cut_gtf_perFeat(annot_gtf, feat_valid, outdir)
		mygtf = gtf_cut_file
	else:
		mygtf = annot_gtf

	# Create a new JSON for SUMMARY, with all keys completed :
	summ_dict = dict()
	summ_dict["queries"] = queries
	summ_dict["priority"] = priority
	summ_dict["gtf"] = annot_gtf
	summ_dict["bed"] = peaks_bed

	with open(outdir+"summary_config.json", "w") as fj:
		json.dump(summ_dict, fj, indent=4)

	distances = reduce(list.__add__, [q["distance"] for q in queries])
	max_distance = int(max(distances))

	logger.info("Distance for peak enlargment: %s", max_distance)

	gtf_index = mygtf + '.sorted.gz'

	if os.path.exists(gtf_index):
		logger.warning(
			"GTF-Index already existed before indexing. Will overwrite old index.")
	try:
		ovls.tabix_index(mygtf)  # Sort, zip, index the gtf
	except IOError:
		logger.error("Unable to create index for file %s. Make sure tabix, bgzip and sort are available and UROPA has rights to (over)write files.", mygtf)
		sys.exit()

	# Validate priority parameter
	pr = True if priority in ['T', 'True', 'TRUE', 'Yes', 'YES', 'Y', 'yes'] else False

	# Create list of Attributes from all queries together,to be shown in final
	# Table-columns for All queries
	query_attributes_none = list([q["show.attributes"] for q in queries])
	query_attributes = [a for a in np.unique(list(chain.from_iterable(query_attributes_none))) if a is not None and a != 'None']

	logger.info("Additional attributes included in output: %s", ','.join(query_attributes))

	# query included, when there is no hit from tabix
	nas_len = len(query_attributes) + 10
	NAsList = list(np.repeat("NA", nas_len))
	NAsList[nas_len - 1] = "- "

	header_base = [
		"peak_id",
		"peak_chr",
		"peak_start",
		"peak_center",
		"peak_end",
		"feature",
		"feat_start",
		"feat_end",
		"feat_strand",
		"feat_anchor",
		"distance",
		"genomic_location",
		"feat_ovl_peak",
		"peak_ovl_feat"]

	header = "\t".join(header_base + query_attributes + ["query"])

	#
	# Preparation of multiprocessing
	#
	spl_dir = outdir + "split_peaks/"
	if args.threads > 1:
		logger.info("Multiprocessing: Peak file will be split in %s smaller files.", args.threads)
		if not os.path.exists(spl_dir):
			os.makedirs(spl_dir)
		cmd = ['split',
			   '-n l/' + str(args.threads),
			   peaks_bed,
			   spl_dir + 'spl_peak_']

		try:
			sp.check_call(cmd, stderr=open(os.devnull, 'w'))
		except sp.CalledProcessError:
			args.threads = 1
			logger.warning(
				"Unable to split peak input into smaller files. Falling back to one thread.")
			logger.info("Check if split command is installed in version 8.22 or higher.")
		except OSError as e:
			args.threads = 1
			logger.warning(
				"Split command not available. Falling back to one thread.")

#
# Processing peaks
#
	input_args = [outdir, gtf_index, query_attributes, queries, max_distance, pr, gtf_has_chr]

	# > Write output according to Thread option
	if args.threads > 1:
		pool = mp.Pool(args.threads)
		partial_func = partial(ant.annotation_process, input_args)
		pool.map(partial_func, glob.glob(spl_dir + "spl_peak_*"))
		pool.close()
		pool.join()
	else:
		ant.annotation_process(input_args, peaks_bed, logger)
		# Files created after annot.process:

	allhits_partials = glob.glob(outdir + "allhits_part_*")
	finalhits_partials = glob.glob(outdir + "finalhits_part_*")

	logger.info("Writing output files to {}".format(outdir))

	if args.add_comments:
		comments = ovls.concat_comments(queries, str(pr), annot_gtf, peaks_bed)
	else:
		comments = None

	#
	# Merging output files
	#
	allhits_outfile = outdir + "allhits.txt"

	if len(queries) > 1 and not pr:
		besthits_outfile = outdir + "besthits.txt"
		merged_outfile = outdir + "finalhits.txt"

		besthits_partials = glob.glob(outdir + "besthits_part_*")
		ovls.finalize_file(merged_outfile, finalhits_partials, header, comments, log=logger)
		ovls.finalize_file(besthits_outfile, besthits_partials, header, comments, log=logger)
	else:
		besthits_outfile = outdir + "finalhits.txt"
		ovls.finalize_file(besthits_outfile, finalhits_partials, header, comments, log=logger)

	logger.debug("Filenames for output files are: {}, {}". format(allhits_outfile, besthits_outfile))

	ovls.finalize_file(allhits_outfile, allhits_partials, header, comments, log=logger)

	#
	# Reformat output
	#
	if args.reformat and len(queries) > 1 and not pr:
		logger.info("Reformatting output...")
		R_reform_Best = [
			'uropa_reformat_output.R',
			'-i',
			besthits_outfile,
			'-k',
			'peak_id',
			'-c',
			'1:5',
			'-d',
			',',
			'-t',
			str(args.threads)]
		try:
			# creates output of Best and gives name "Reformatted_"
			logger.debug('Reformat output call is {}'.format(R_reform_Best))
			pr_R = sp.check_output(R_reform_Best)
		except sp.CalledProcessError:
			logger.warning("Reformatted output could not be created with call %s", ' '.join(R_reform_Best))
		except OSError:
			logger.warning("Rscript command not available for summary output.")

	#
	# Visual summary
	#
	if args.summary:
		logger.info("Creating the Summary graphs of the results...")
		summary_script = "uropa_summary.R"
		summary_output = outdir + "summary.pdf"

		if len(queries) > 1 and not pr and os.path.exists(merged_outfile):
			call = [
				summary_script,
				"-f",
				merged_outfile,
				"-c",
				outdir + "summary_config.json",
				"-o",
				summary_output,
				"-b",
				besthits_outfile]
		else:
			call = [
				summary_script,
				"-f",
				besthits_outfile,
				"-c",
				outdir + "summary_config.json",
				"-o",
				summary_output]
		try:
			logger.debug('Summary output call is {}'.format(call))
			sum_pr = sp.check_output(call)
		except sp.CalledProcessError:
			logger.warning("Visualized summary output could not be created from %s.", ' '.join(call))
		except OSError:
			logger.warning("Rscript command not available for summary output.")

	#
	# Clean up of temporary files
	#
	outputs_ready = os.path.exists(
		allhits_outfile) and os.path.exists(besthits_outfile)

	if outputs_ready:
		#logger.debug("Attempting to clean {}".format(outdir))
		ovls.cleanup(outdir, logger)

		if os.path.exists(spl_dir):
			shutil.rmtree(spl_dir)
		if os.path.exists(outdir+"summary_config.json"):
			os.remove(outdir+"summary_config.json")
		os.remove(gtf_index)  # .gz
		os.remove(gtf_index + ".tbi")
		if len(gtf_feat) > 1:
			os.remove(gtf_cut_file)
			os.remove(gtf_cut_file + ".sorted")

	logger.info("End time: %s", datetime.datetime.now().strftime("%d.%m.%Y %H:%M"))



"""Contains code for annotation process."""
import sys
import operator
import numpy as np
import pysam
from . import overlaps as ovls


def annotation_process(input_args, peak_file, log=None):
    """Contains main logic for the annotation process."""
    [outdir, gtf_index, attrib_k, queries, max_distance,
     priority, gtf_has_chr] = input_args

    try:
        anno = pysam.TabixFile(gtf_index)
    except IOError:
        log.error("Unable to open tabix index {}.".format(gtf_index))
    except ValueError:
        log.error("Tabix index file is missing ({})".format(gtf_index))

    try:
        prefix_pk = peak_file.split("_peak_")[1]
    except IndexError:
        # In case peaks file is small and not splitted
        prefix_pk = ""

    with open(peak_file, 'r') as peaks:
        Best_hits_tab = dict()
        All_hits_tab = dict()
        min_dist = dict()

        from collections import OrderedDict
        for j in range(len(queries)):
            min_dist[j] = dict()
            Best_hits_tab[j] = OrderedDict()
            All_hits_tab[j] = OrderedDict()

        counter = 0
        for p in peaks:
            counter += 1

            peak = ovls.parse_peak(p, extend=max_distance)
            if peak is None:
                if log is not None:
                    log.warning("Skipping invalid peak '{}'".format(p))
                continue
            #peak['id'] = peak['name'] if peak[
            #    'name'] is not None else peak['id']
            # Re-initialise table records for each peak
            for q in enumerate(queries):
                All_hits_tab[q[0]][peak['id']] = ""
                Best_hits_tab[q[0]][peak['id']] = ""

                q[1]["distance"] = [abs(int(d)) for d in q[1]["distance"]]
                if len(q[1]["distance"]) == 2:
                    min_dist[q[0]][peak['id']] = [q[1]["distance"], list()]
                else:
                    min_dist[q[0]][peak['id']] = [q[1]["distance"][0], list()]

            # Run TABIX
            chrom_db = ovls.peak_has_chr(peak['chr'], gtf_has_chr)
            tabix_query = chrom_db + ":" + \
                str(peak['estart']) + '-' + str(peak['eend'])

            try:
                hits = anno.fetch(tabix_query, multiple_iterators=True)
            except ValueError:
                hits = list()
                if log is not None:
                    log.warning("Region {} could not be found in GTF annotation for peak {}. Tabix query was: {}:{}-{}".format(
                        chrom_db, peak['id'], chrom_db, str(peak['estart']), str(peak['eend'])))

            has_hits = list()
            # keep array of  all valid queries (T,F) for all hits in the peak.
            valid_q_per_peak = list()
            dict_vqp = dict()

            # Collection and analysis of all hits for the peak
            for hit in hits:  # All overlapping features
                hit = hit.split("\t")  # from "anno.fetch" hit is a string
                h = {'feature': hit[2], 'strand': hit[6]}
                hit_center = np.mean([int(hit[3]), int(hit[4])])

                dist_from_peak = min(
                    abs(int(peak['estart']) - int(hit[3])),
                    abs(int(peak['start']) - int(hit[3])),
                    abs(int(peak['eend']) - int(hit[3])),
                    abs(int(peak['end']) - int(hit[3])),
                    abs(int(peak['eend']) - int(hit[4])),
                    abs(int(peak['end']) - int(hit[4])),
                    abs(int(peak['estart']) - int(hit[4])),
                    abs(int(peak['start']) - int(hit[4])),
                    abs(peak['center'] - hit_center),
                    abs(peak['center'] - int(hit[3])),
                    abs(peak['center'] - int(hit[4])))
                # Verify that hit is valid at least for one of the distances of
                # queries
                valid_dist = any([dist_from_peak <= int(
                    max(q["distance"])) for q in queries])
                # Keep hit if internals are required in any query
                i_want_internals = any([q["internals"] in [
                    ['T'], ['True'], ['TRUE'], ['Yes'], ['YES'], ['Y'], ['yes'], ['F'], ['False'], ['FALSE'], ['No'], ['NO'], ['N'], ['no']] for q in queries])

                if (i_want_internals) and not valid_dist:
                    # Keep if gene inside the peak even if Dist > config.D
                    feat_in_peak = (int(peak['start']) < int(
                        hit[3]) and int(hit[4]) < int(peak['end']))
                    peak_in_feat = (int(hit[3]) < int(
                        peak['start']) and int(peak['end']) < int(hit[4]))

                    if feat_in_peak or peak_in_feat:
                        valid_dist = True

                # Find hits with valid values for the queries (Search all
                # queries ,Not only PRIORITY)
                v_fsa = [ovls.valid_fsa(h, hit, q, peak['strand']) for q in queries]

                # Pair the Valid query values(fsb) with valid_distance and
                # valid strand for each query
                vsd = [[x, valid_dist] for x in v_fsa]
                valid_queries = [i for i, v in enumerate(vsd) if all(v)]

                hitj = "\t".join(hit)
                # Create dictionary of hit with its valid query per peak
                if valid_queries:
                    has_hits.append(True)
                    valid_q_per_peak.append(valid_queries)
                    dict_vqp.update({hitj: valid_queries})

                if not valid_queries:
                    has_hits.append(False)

            # ----- All hits parsed. Check only hits with valid-queries now ----- #
            #log.debug("\n---All hits parsed for the peak.Totally, the Peak has {} hits from which {} are Valid ---".format( len(has_hits), has_hits.count(True) ))
            # 9 cols= feat, f.start, f.end, f.strand, dist, min_pos, genom_loc,
            # feat-to-peak-ovl , peak-to-feat-ovl (no Q)
            nas_len = len(attrib_k) + 9
            NAsList_q = list(np.repeat("NA", nas_len))

            # Search hits with only Prior or Secondary Queries
            for hitj, vq in list(dict_vqp.items()):
                hit = hitj.split("\t")
                hit_len = abs(int(hit[4]) - int(hit[3]))
                strand = hit[6]
                for j in vq:
                    # Check DIRECTION before Closest_Distance, if given
                    if queries[j]["direction"] != ['any_direction']:
                        correct_dir = ovls.define_direction(queries[j]["direction"], int(hit[3]), int(
                            hit[4]), strand, peak['length'], int(peak['start']), int(peak['end']), peak['center'])
                        #logger.debug("\nFor the hit {}-{}, direction is correct : {}".format(hit[3],hit[4] ,correct_dir))

                        if correct_dir:
                            # IF direction of hit is the desired, then filter
                            # for Distance
                            min_pos, Dhit = ovls.distance_to_peak_center(peak['center'], int(
                                hit[3]), int(hit[4]), strand, queries[j]["feature.anchor"])
                            #logger.debug("\nClosest Distance from the peak center : {} ".format(Dhit))

                        if not correct_dir and Best_hits_tab[j][peak['id']] == "" and All_hits_tab[j][peak['id']] == "":
                            # hit with Incorrect Dir: move to new peak, fill in NAs if  peak is empty
                            #log.debug("\nThe hit doesn't have the correct direction -> NAs in All & Best hits table -> Continue to New hit.")
                            All_hits_tab[j][peak['id']] = "\t".join(np.hstack([peak['name'], peak['chr'], peak[
                                'start'], peak['center'], peak['end'], NAsList_q, str(j) + "\n"]))
                            Best_hits_tab[j][peak['id']] = "\t".join(np.hstack([peak['name'], peak['chr'], peak[
                                'start'], peak['center'], peak['end'], NAsList_q, str(j) + "\n"]))
                            continue  # Move to next valid query and then next hit

                        # When direction NOT correct but tables not empty.
                        elif not correct_dir and Best_hits_tab[j][peak['id']] != "" and All_hits_tab[j][peak['id']] != "":
                            continue

                    elif queries[j]['direction'] == ['any_direction']:
                        min_pos, Dhit = ovls.distance_to_peak_center(peak['center'], int(
                            hit[3]), int(hit[4]), strand, queries[j]["feature.anchor"])
                        #log.debug("\nClosest Distance(Dhit) when No Direction defined :{} ".format(Dhit))

                    # > Check OVERLAPS of peak-hit, define genomic_location
                    genomic_location = "not.specified"
                    genomic_location, ovl_pk, ovl_feat = ovls.overlap_peak_feature(genomic_location, int(
                        peak['start']), peak['center'], int(peak['end']), peak['length'], hit_len, int(hit[3]), int(hit[4]), strand)

                    # > After Direction is checked move on to check Distance hit <--> peak
                    same_gene = all(x in min_dist[j][peak['id']][1] for x in [hit[2], hit[3], hit[
                        4], hit[6]])  # next hit shouldn't have same {feat,start,end}
                    internals_location = list()
                    if "Inside" in genomic_location:
                        # > Check internals direction for defining Bi-direct. distance window
                        internals_location = [ovls.find_internals_dir(a, peak['center'], int(hit[3]), int(hit[4]), strand) for a in queries[j]["feature.anchor"]]

                    if len(queries[j]["distance"]) == 2:
                        d_is_best = ovls.get_distance_by_dir([min_dist[j][peak['id']][0][0], min_dist[
                            j][peak['id']][0][1]], genomic_location, internals_location, Dhit, queries[j]["internals"])
                    if len(queries[j]["distance"]) == 1:
                        d_is_best = ovls.get_distance_by_dir([min_dist[j][peak['id']][0], min_dist[j][
                            peak['id']][0]], genomic_location, internals_location, Dhit, queries[j]["internals"])
                        #log.debug('Called with {} (q {}): {}, {}, {}, {}'.format(peak['id'], j, min_dist[j][peak['id']][0], genomic_location, internals_location, Dhit))
                        #log.debug('Returned {} {}'.format(peak['id'], d_is_best))

                    # print "Best distance found :{}".format(d_is_best)
                    if d_is_best and not same_gene:  # Dhit < Dbest
                        best_hit_dist = min_dist[j][peak['id']][0]
                        current_overlap = ovls.get_besthit(j, len(queries[j]["distance"]), peak['id'], hit, attrib_k, Dhit)
                        #log.debug("best_hit_dist is {}".format(best_hit_dist))
                        #log.debug("current_overlap dist is {}".format(current_overlap[0]))

                        if current_overlap[0] <= best_hit_dist:
                            #log.debug("I just confirmed that {} is smaller than {}, or internals T".format(current_overlap[0], best_hit_dist))
                            min_dist[j][peak['id']] = current_overlap
                            Best_res = ovls.create_table(peak['name'], peak['chr'], peak['start'], peak['end'], str(peak['center']), min_dist[j][peak['id']], attrib_k, min_pos, genomic_location, ovl_pk, ovl_feat, j)
                            Best_hits_tab[j][peak['id']] = Best_res

                    # Dhit > Dbest
                    # only if empty fill in with NAs
                    elif not d_is_best and not same_gene and Best_hits_tab[j][peak['id']] == "":
                        #log.debug("\nDistance of Hit from Peak Center = {}, LARGER THAN current Min Distance = {}\n".format(Dhit, min_dist[j][peak['id']][0]))
                        Best_hits_tab[j][peak['id']] = "\t".join(np.hstack([peak['name'], peak['chr'], peak[
                            'start'], str(peak['center']), peak['end'], NAsList_q, str(j) + "\n"]))
                    #log.debug("min_dist is {}".format(min_dist[j][peak['id']]))

                    # EVERY HIT getting registered in Allhits file if Dhit <
                    # query.distance
                    if len(queries[j]["distance"]) > 1:
                        Dhit_smaller = ovls.get_distance_by_dir([queries[j]["distance"][0], queries[j][
                            "distance"][1]], genomic_location, internals_location, Dhit, queries[j]["internals"])
                    else:
                        Dhit_smaller = ovls.get_distance_by_dir([queries[j]["distance"][0], queries[j][
                            "distance"][0]], genomic_location, internals_location, Dhit, queries[j]["internals"])

                    if Dhit_smaller and not same_gene:
                        #log.debug("Distance of hit SMALLER than Config Distance -> Hit written to All_Hits.")
                        All_hits_tab[j][peak['id']] = ovls.write_hit_to_All(All_hits_tab, peak['id'], attrib_k, Dhit, hit, peak['name'], peak['chr'], peak['start'],
                                                                            peak['end'], str(peak['center']), min_pos, genomic_location, ovl_pk, ovl_feat, j)

                    # Hit not fitting in any of the distance values
                    if not Dhit_smaller and queries[j]["internals"] != ["True"] and (All_hits_tab[j][peak['id']] == ''):
                        #log.debug("\nDhit= {} is LARGER than config Distance:{}.".format(Dhit, queries[j]["distance"]))
                        # Write over empty or NA the new NA, for new hit,so
                        # each hit is parsed.
                        All_hits_tab[j][peak['id']] = "\t".join(np.hstack([peak['name'], peak['chr'], peak[
                            'start'], str(peak['center']), peak['end'], NAsList_q, str(j) + "\n"]))

                    # Hits with further Distance -> Check for Internals
                    if not Dhit_smaller and queries[j]["internals"] == ["True"] and not same_gene:
                        if "Inside" in genomic_location:
                            #log.debug("\n-> Hit is 'Internal' with distance LARGER than config.distance.It will be recorded to the All_hits table.")
                            All_hits_tab[j][peak['id']] = ovls.write_hit_to_All(All_hits_tab, peak['id'], attrib_k, Dhit, hit, peak['name'], peak['chr'], peak['start'], peak['end'], str(peak['center']),
                                                                                min_pos, genomic_location, ovl_pk, ovl_feat, j)

                    if not Dhit_smaller and All_hits_tab[j][peak['id']] != '':
                        #log.debug("Hit has larger distance than allowed ")
                        continue

            # > Add to Best Hits the Best Internal features(when internals=True, D>config."distance"),
            # after having filled-in the All Hits for the peak.
            for hitj, vq in list(dict_vqp.items()):
                for j in vq:
                    if Best_hits_tab[j][peak['id']] != "" and All_hits_tab[j][peak['id']] != "":
                        if (Best_hits_tab[j][peak['id']].split("\t")[10] == "NA" and
                                All_hits_tab[j][peak['id']].split("\t")[10] != "NA" and
                                queries[j]["internals"] == ["True"]):

                            hit_line = All_hits_tab[j][peak['id']].split("\n")
                            hit_line = [h for h in hit_line if h != '']
                            internal = [h.split("\t")[11] for h in hit_line]
                            hit_dist = [int(
                                h.split("\t")[10]) for h in hit_line]
                            mv_pos, min_val = min(
                                enumerate(hit_dist), key=operator.itemgetter(1))

                            if internal[mv_pos] in ["PeakInsideFeature", "FeatureInsidePeak"] and min_val != []:
                                Best_hits_tab[j][peak['id']] = hit_line[
                                    mv_pos] + "\n"
                                #logger.debug("\nBest Internal hit to be recorded: {}\n".format(hit_line[mv_pos]+"\n"))

                    else:  # if All_hits_tab[j][peak['id']] == "" :
                        continue

            # When no hits are valid for one query-> All_hits[q] stays <""> ->
            # Get them filled with NAs
            for a in range(len(queries)):
                if All_hits_tab[a][peak['id']] == "":
                    #log.debug("\nQuery {} didn't validate any hit ! All_hits will be filled in with NAs".format(a))
                    All_hits_tab[a][peak['id']] = "\t".join(np.hstack([peak['name'], peak['chr'], peak[
                        'start'], str(peak['center']), peak['end'], NAsList_q, str(a) + "\n"]))
                if Best_hits_tab[a][peak['id']] == "":
                    #log.debug("\nQuery {} didn't validate any hit !Best hits will be filled in with NAs".format(a))
                    Best_hits_tab[a][peak['id']] = "\t".join(np.hstack([peak['name'], peak['chr'], peak[
                        'start'], str(peak['center']), peak['end'], NAsList_q, str(a) + "\n"]))

            # end.for_hit
            if not has_hits or True not in has_hits:
                # Write the Non Overlaps in the first dict,so it is saved only once.
                # log.debug ("\nPeak has no hits ! All queries Tables will be
                # filled in with NAs\n") #only to [0]
                for j in range(len(queries)):
                    All_hits_tab[j][peak['id']] = "\t".join(np.hstack([peak['name'], peak['chr'], peak[
                        'start'], str(peak['center']), peak['end'], NAsList_q, str(j) + "\n"]))
                    Best_hits_tab[j][peak['id']] = "\t".join(np.hstack([peak['name'], peak['chr'], peak[
                        'start'], str(peak['center']), peak['end'], NAsList_q, str(j) + "\n"]))

            # > PRIORITY FLIPPING
            # Check for each Line in the Tabs if Query 1 = NA, when Pr= True
            # and replace with Secondary Query.
            if priority and len(queries) > 1:
                # empty if only 0 exists
                rest_q = [j for j in range(len(queries)) if j != 0]
                # ** All_hits_per_peak -> may have lots of hits, sep with \n

                isNA = list()
                for i in enumerate(queries):
                    if All_hits_tab[i[0]][peak['id']]:
                        isNA.append(All_hits_tab[i[0]][
                            peak['id']].split("\t")[9] == "NA")

                QnotNA = [pos for pos, val in enumerate(isNA) if val is False]
                Q_NA = [pos for pos, val in enumerate(isNA) if val is True]

                # All have Hit ,and Priority query
                if isNA[0] is False or not any(isNA):
                    for r in rest_q:
                        All_hits_tab[r][peak['id']] = Best_hits_tab[
                            r][peak['id']] = ""

                # 1st Query is NA, there are Sec.Queries. // any(QnotNA) = []
                # when isNA=[True] for one Q
                if isNA[0] is True and any(QnotNA):
                    # Queries with NA: map(lambda qna :
                    # All_hits_tab[qna][peak['id']], Q_NA)
                    repl_q = min(QnotNA)  # The query to be used instead of qO.
                    ##log.debug ("Priority Query hasn't given any hit, so it will be replaced by secondary Query's Hit: {} ".format (repl_q))
                    All_hits_tab[0][peak['id']] = Best_hits_tab[
                        0][peak['id']] = ""
                    ##log.debug ("\nHits from Secondary Q to replace priority query: \n{} ".format(All_hits_tab[repl_q][peak['id']] ))
                    rest_QnotNA = [qn for qn in QnotNA if qn != repl_q]
                    if any(rest_QnotNA):
                        for qr in rest_QnotNA:
                            All_hits_tab[qr][peak['id']] = Best_hits_tab[
                                qr][peak['id']] = ""
                    for qna in Q_NA:
                        All_hits_tab[qna][peak['id']] = Best_hits_tab[
                            qna][peak['id']] = ""

                if all(isNA):
                    #logg.debug("\nNo query has any Hit.No replacement of Priority Query possible-> NAs will be filled in the Output.")
                    TabInList_p = [All_hits_tab[l][
                        peak['id']] for l in range(len(queries))]
                    ##log.debug("Hit lines for all queries are : {}".format(TabInList_p ))
                    # If all_hits_tab doesn't have all queries will have ""
                    TabInList_p = [Tab for Tab in TabInList_p if Tab != ""]

                    for j in enumerate(TabInList_p):
                        if j[1] == "":
                            TabInList_p[j[0]] = "\t".join(np.hstack([peak['name'], peak['chr'], peak[
                                'start'], str(peak['center']), peak['end'], NAsList_q, str(j[0]) + "\n"]))

                    All_hits_tab[0][
                        peak['id']] = ovls.merge_queries(TabInList_p)
                    Best_hits_tab[0][
                        peak['id']] = ovls.merge_queries(TabInList_p)
                    ##log.debug("\nQueries hits for this peak to be merged in the Q.col:{}".format(ovls.merge_queries(TabInList_p)) )
                    for r in rest_q:
                        All_hits_tab[r][peak['id']] = Best_hits_tab[
                            r][peak['id']] = ""

        All_combo = OrderedDict()  # Output peaks with same order as All_hits_tab
        mydict = All_hits_tab
        for k in mydict[0].keys():
            All_combo[k] = [pid[k] for q, pid in list(mydict.items())]

        #  Best hits  #
        Best_combo = OrderedDict()
        mybestD = Best_hits_tab
        for k in mybestD[0].keys():
            Best_combo[k] = [pid[k] for q, pid in list(mybestD.items())]

        def all_same(items):
            """Returns true if all items of a list are the same."""
            return all(x == items[0] for x in items)

        # > Merge hits of queries to have one best, if queries >1
        if len(queries) > 1 and not priority:
            BestBest_hits = OrderedDict()
            for k in Best_combo:
                # [  [] , [] , [] ]-> can be of same query, same distance
                records = [s.split("\n") for s in Best_combo[k]]
                # split also internally each query's string to see if it
                # contains more >1 hits(when same distance, more >1 are conc.
                # to same query)
                spl_rec = [[t if t != '' else None for t in r] for r in records]
                recs = [x + "\n" for s in spl_rec for x in s if x != None]

                if len(recs) > 1:  # Only when more than two hits, and no NAs here
                    splitted_hits = [recs[h].split("\t") for h in range(len(recs))]
                    # s= each hit line in string
                    splitted_hits = [s for s in splitted_hits if s != [""]]
                    featOfHit = [splitted_hits[s][
                        5] for s in range(len(splitted_hits))]
                    startPos = [splitted_hits[s][
                        6] for s in range(len(splitted_hits))]
                    endPos = [splitted_hits[s][
                        7] for s in range(len(splitted_hits))]
                    Dist_hits = [float(splitted_hits[s][10]) if splitted_hits[s][
                        10] != "NA" else float("Inf") for s in range(len(splitted_hits))]  # 9th col= Distance
                    # Dist_hit = [Dist_hits[0] if all_same(Dist_hits) else
                    # Dist_hits]  #Either same dist or All NAs

                    # When All "NAs" for all queries-> keep only line from 1st
                    # query
                    NA_hits = ["NA" if splitted_hits[s][
                        10] == "NA" else None for s in range(len(splitted_hits))]
                    # If None means ALL queries have a feature
                    all_have_feat = all(x is None for x in NA_hits)

                    same_feat = (all_same(featOfHit) and all_same(
                        startPos) and all_same(endPos) and all_same(Dist_hits))

                    if len(splitted_hits) > 1:
                        if all_same(NA_hits):
                            # BestBest_hits[k] =  Best_combo[k][0]  #Keep only
                            # one of the hits wth NA
                            BestBest_hits[k] = ovls.merge_queries(
                                Best_combo[k])

                        if (all_have_feat and same_feat) or (not same_feat and all_same(Dist_hits)):
                            BestBest_hits[k] = ovls.merge_queries(
                                Best_combo[k])

                        if (all_have_feat) and not same_feat or (not all_same(NA_hits) and not all_have_feat):
                            # min_pos,min_v = min(enumerate(Dist_hits),
                            # key=operator.itemgetter(1)) ## If only one min in
                            # one pos
                            min_v = min(Dist_hits)
                            # retrieve the value of the list with [0],should be
                            # only 1 !
                            min_pos = [min_pos for min_pos, v in enumerate(Dist_hits) if v == min_v][
                                0]
                            # Keep to BB the pos with Min_Dist
                            BestBest_hits[k] = recs[min_pos]

                    elif len(splitted_hits) == 1:
                        BestBest_hits[k] = Best_combo[k]

        # > Write in file
        allhits_file = outdir + "allhits_part_" + prefix_pk + ".txt"

        try:
            ovls.write_partial_file(allhits_file, All_combo)
        except IOError:
            if not log is None:
                log.error("Unable to open file " + allhits_file + " for writing results!")
            sys.exit()

        finalhits_file = outdir + "finalhits_part_" + prefix_pk + ".txt"
        if len(queries) > 1 and not priority:
            besthits_file = outdir + "besthits_part_" + prefix_pk + ".txt"
            ovls.write_partial_file(besthits_file, Best_combo)
            ovls.write_partial_file(finalhits_file, BestBest_hits)
        else:
            ovls.write_partial_file(finalhits_file, Best_combo)
    return



#!/usr/bin/env python
"""Contains functions for UROPA overlap evaluation."""

import os
import re
import sys
import subprocess
import logging
import shlex
import shutil
import glob
import numpy as np


def peak_has_chr(chrom, gtf_has_chr):
    """Checks if the peak chromosome name against the reference chromosome name."""
    p_wo_chr = "chr" not in chrom
    gtf_no_chr = not gtf_has_chr

    if p_wo_chr and gtf_has_chr:  # Peak 'chr' should have chr if gtf has
        chrom_db = "chr" + chrom

    elif not p_wo_chr and not gtf_has_chr:  # peak has chr, gtf doesn't-> remove  chr from peak
        chrom_db = chrom.strip("chr")

    elif (gtf_no_chr and p_wo_chr) or (gtf_has_chr and not p_wo_chr):  # none has chr -> keep p chrom
        chrom_db = chrom

    return chrom_db


def check4_chr(file):
    """Checks if a file has lines starting with 'chr'."""
    if os.path.exists(file):
        with open(file, "r") as f:
            flines = f.readlines()

        comm_lines = len([li for li in flines if li.startswith("#")])
        first_line = flines[comm_lines]
        file_has_chr = first_line.startswith("chr")  # >> True / False
        return file_has_chr


def tabix_index(annot_gtf):
    """ Prepares gtf for query: Performs sorting, zipping and indexing of given gtf"""
    out_sorted = annot_gtf + ".sorted"
    out_zipped = out_sorted + ".gz"

    with open(out_zipped, 'w'):
        os.system('grep  -v  ^#  ' + annot_gtf +
                  '  |  sort -k1,1 -k4,4n > ' + annot_gtf + '.sorted')
        os.system('bgzip -c  -f ' + annot_gtf +
                  '.sorted ' + ' > ' + out_zipped)

    run_tabix = 'tabix -f  ' + out_zipped
    idx = subprocess.Popen(shlex.split(run_tabix), stdout=subprocess.PIPE)
    idx.wait()


def valid_fsa(h, hit, q, pstrand):
    """Returns if a hit h is valid for query q in their basic common keys."""

    vf = (h["feature"] == q["feature"][0])
    v_str = valid_strand(h["strand"], pstrand, q["strand"][0])
    va = valid_attribute(q["filter.attribute"][0], q["attribute.value"][0], hit)
    return all([vf, v_str, va])


def valid_strand(hit_str, p_str, q_str):
    """Returns boolean showing match of peak.strand to feature.strand according to query requirements """
    if p_str != ".":
        if q_str == "ignore":
            return True if (hit_str != p_str) or (hit_str == p_str) else False
        if q_str == "same":
            return True if (hit_str == p_str) else False
        if q_str == "opposite":
            return True if (hit_str == "+" and p_str == "-") or (hit_str == "-" and p_str == "+") else False
            # OR : return (True if (hit_str != p_str) ,i.e when one is "." and
            # other is "+,-"

    elif p_str == "." or hit_str == ".":
        return True


def valid_attribute(attr_filter_key, attr_filter_val, hit):
    """Validates the hit according to a filter attribute."""
    if (attr_filter_key != "None") and (attr_filter_val != "None"):
        try:
            # If key for filtering is not correct or doesn't exist-> error
            # should be ignored
            hit_attrib_val = re.split(
                "; " + attr_filter_key + " ", hit[8])[1].split(';')[0].strip('"\'').rstrip('\"')
        except IndexError:  # if key doesn't exist re.split will give error
            hit_attrib_val = "NA"

        # If biotype of hit == attr_value from query-> continue annotation
        return attr_filter_val == hit_attrib_val
    else:
        return True  # no filtering given->ignore


def file_len(fname):
    """Returns Number of lines of a file"""
    with open(fname) as f:
        i = -1
        for i, l in enumerate(f):
            pass
    return i + 1


def parse_peak(peakstr, extend=0, delim='\t'):
    """ Parses a string into peak dictionaries """
    fields = peakstr.replace('\n', '').replace('\r', '').split(delim)
    num_fields = len(fields)

    if num_fields < 3:
        return None

    p = dict()
    if num_fields >= 3:
        p['chr'] = fields[0]
        p['start'] = fields[1]
        p['end'] = fields[2]

    if num_fields >= 4:
        if fields[3] != '.':
            p['name'] = fields[3]

    if num_fields >= 5:
        p['score'] = fields[4]

    if num_fields >= 6:
        p['strand'] = fields[5]

    defaults = {'id': p['chr'] + ":" + p['start'] + "-" + p['end'], 'name': p['chr'] + ":" + p['start'] + "-" + p['end'], 'score': 0, 'strand': ".",
                'center': int(np.around(np.mean([int(p['start']), int(p['end'])]))), 'estart': max(1, int(p['start']) - extend), 'eend': int(p['end']) + extend, 'length': abs(int(p['end']) - int(p['start']))}
    if 'name' in list(p.keys()):
        p['id'] = p['chr'] + ":" + p['start'] + "-" + p['end'] + "_" + p['name']

    values = [p[k] if k in p else defaults[k] for k in list(defaults.keys())]
    peak = dict(list(zip(list(p.keys()) + list(defaults.keys()), list(p.values()) + values)))
    return peak


def define_direction(dir_key, fstart, fend, strand, p_len, pstart, pend, p_center):
    """Verifies if direction of peak is the same as required by query"""

    peak_dir = find_peak_dir(fstart, fend, strand, pstart, p_center, pend)
# Inverse in case of negative strand
    feat_start = fstart if strand == "+" else fend

# How much internally to peak can a feature be, to be labeled as upstr/downstr.
    ten_percent = p_len * 0.1

    if abs(p_center - feat_start) >= ten_percent:
        if peak_dir == "upstream" and dir_key == ["upstream"]:
            return True
        return peak_dir == "downstream" and dir_key == ["downstream"]

    if abs(p_center - feat_start) < ten_percent:
        return False


def distance_to_peak_center(p_center, hit3, hit4, strand, feat_pos):
    """Returns the position of feature from which the distance is minimum to the peak and the measured distance to peak center"""
    import operator

    feat_start = hit3 if strand == "+" else hit4
    feat_end = hit4 if strand == "+" else hit3

    hit_center = np.mean([feat_start, feat_end])
    hit_pos = {"start": feat_start, "center": hit_center, "end": feat_end}
    d_pos = [abs(hit_pos[i] - p_center) for i in feat_pos]
    pos, dmin = min(enumerate(d_pos), key=operator.itemgetter(1))

    min_pos = feat_pos[pos]
    return min_pos, dmin


def find_peak_dir(fstart, fend, strand, pstart, p_center, pend):
    """ Returns the location of the  queried peak relative to the feature direction """
    feat_start = fstart if strand == "+" else fend
    feat_end = fend if strand == "+" else fstart

    # Add one base to single-base features to give length for direction
    # detection.
    if feat_start == feat_end and strand == "+":
        feat_end = feat_start + 1
    elif feat_start == feat_end and strand == "-":
        feat_start = feat_end + 1

    up_sides = abs(p_center - feat_start) < abs(p_center -
                                                feat_end)  # left and right
    down_sides = abs(p_center - feat_start) > abs(p_center - feat_end)

    upstream = ((abs(pstart - feat_end) > abs(pstart - feat_start))
                and (abs(pend - feat_end) > abs(pend - feat_start)))
    downstream = ((abs(pstart - feat_end) < abs(pstart - feat_start))
                  and (abs(pend - feat_end) < abs(pend - feat_start)))

    if upstream and up_sides:
        return "upstream"
    elif downstream and down_sides:
        return "downstream"
    else:
        return "not.specified"


def find_internals_dir(anchor, p_center, feat_start, feat_end, strand):
    """ Determine the direction of peak when feature-inside-peak or peak-inside-feature """
    # +strand: When internal feature/peak is Left of p.center, dir->downstream, otherwise dir-> upstream,
    # -strand :inverse of above
    feat_center = np.mean([feat_start, feat_end])
    feat_pos = {"start": feat_start, "center": feat_center, "end": feat_end}
    if strand == "+":
        return "downstream" if feat_pos[anchor] < p_center else "upstream"

    if strand == "-":
        return "upstream" if feat_pos[anchor] < p_center else "downstream"


def get_distance_by_dir(inputD, genom_loc, intern_loc, Dhit, internals_allowed=["False"]):
    """Limits the hit by a distance window depending on peak's position relative-to-feature """

    [D_upstr, D_downstr] = inputD

    if genom_loc == "upstream" or genom_loc == "overlapStart":
        return Dhit <= D_upstr
    if genom_loc == "downstream" or genom_loc == "overlapEnd":
        return Dhit <= D_downstr

    # Rescue if internal peak
    if internals_allowed == ["True"]:
        return(any(intern_loc))
    else:
        best_dist = [Dhit <= D_upstr if l == "upstream" else Dhit <= D_downstr for l in intern_loc]
        return(any(best_dist))


def detect_internals(pstart, pend, hit_start, hit_end):
    """[Local] Check for internal features (for 'genomic_location' column, independent of users key option) """

    feat_in_peak = (pstart < hit_start and hit_end < pend)
    peak_in_feat = (hit_start < pstart and pend < hit_end)

    if feat_in_peak:
        return "FeatureInsidePeak"
    elif peak_in_feat:
        return "PeakInsideFeature"
    else:
        return "not.specified"


def round_up(val):
    """[Local] Rounds a value on second decimal """
    import math
    val = math.ceil(val * 100) / 100
    return val


# import "division" allows decimals
def calculate_overlap(pstart, pend, peakL, featL, hit_start, hit_end):
    """[Local] Returns value of overlap btw {0,1} representing percentage of length covered by the peak and that covered by the feature """
    p_range = list(range(pstart, pend))
    feat_range = list(range(hit_start, hit_end))
    pset = set(p_range)

    ovl_range = pset.intersection(feat_range)
    peakL = 1 if peakL == 0 else peakL
    featL = 1 if featL == 0 else featL

    ovl_pk = round_up(len(ovl_range) / peakL)
    ovl_pk = 1 if ovl_pk > 1 else ovl_pk
    ovl_feat = round_up(len(ovl_range) / featL)
    ovl_feat = 1 if ovl_feat > 1 else ovl_feat

    return ovl_pk, ovl_feat, ovl_range


def define_genom_loc(current_loc, pstart, p_center, pend, hit_start, hit_end, hit_strand, ovl_range):
    """ [Local] Returns location label to be given to the annotated peak, if upstream/downstream or overlapping one edge of feature."""
    all_pos = ["start", "end"]
    closest_pos, dmin = distance_to_peak_center(
        p_center, hit_start, hit_end, hit_strand, all_pos)
    if current_loc == "not.specified":  # Not internal
        if closest_pos == "end" and any(ovl_range):
            return "overlapEnd"
        elif closest_pos == "start" and any(ovl_range):
            return "overlapStart"
        elif not any(ovl_range):
            # Check about direction :"upstream", "downstream"
            current_loc = find_peak_dir(
                hit_start, hit_end, hit_strand, pstart, p_center, pend)
            return current_loc
    return current_loc


def overlap_peak_feature(genom_loc, pstart, p_center, pend, peakL, featL, hit_start, hit_end, hit_strand):
    """Gives the collective information of the location and overlap of a peak to a feature """
    # A/ Find internal features/peaks
    # hit_start, hit_end : as given in gtf/no inversion for '-'strand.
    genom_loc = detect_internals(pstart, pend, hit_start, hit_end)
    # B/ Calculate overlap Ratio peak/feat and feat/peak
    ovl_pk, ovl_feat, ovl_range = calculate_overlap(
        pstart, pend, peakL, featL, hit_start, hit_end)

    # C/ Define genomic_location with or w/o overlap (input is internals
    # genom_loc for comparison and update )
    genom_loc = define_genom_loc(
        genom_loc, pstart, p_center, pend, hit_start, hit_end, hit_strand, ovl_range)
    return genom_loc, ovl_pk, ovl_feat


def get_hit_attribute(hit, attribute):
    """Splits attributes from hit lines"""
    pos_match = [i if re.match(
        attribute, hit_a) else None for i, hit_a in enumerate(hit[8].split("; "))]
    pos_match = [x for x in pos_match if x is not None]
    if len(pos_match) > 0:
        attr_val = [hit[8].split("; ")[m].split(" ")[
            1].strip('"\'').rstrip('\";') for m in pos_match]
        val = [av for av in attr_val if av is not None][0]
        val = val.replace("\t", "").replace("\r", "").replace("\n", "")
        return val
    else:
        return "not.found"


def get_besthit(q, len_q_dist, p_nm, hit, attrib_keys, Dhit):
    """Defines the best hit for a query."""
    Dhit = [Dhit, Dhit] if len_q_dist == 2 else Dhit

    ret = None
    if attrib_keys != ["None"]:
        #attr_val = map(lambda k:  re.split(k+" ", hit[8])[1].split(';')[0].strip('"\'').rstrip('\"'), attrib_keys)
        # list of all values if multiple keys
        attr_val = [get_hit_attribute(hit, a) for a in attrib_keys]
        ret = [Dhit, [hit[2], hit[3], hit[4], hit[6], attr_val]]

    elif attrib_keys == ["None"]:
        ret = [Dhit, [hit[2], hit[3], hit[4], hit[6]]]

    return ret


def create_table(peak_id, chrom, pstart, pend, p_center, min_dist_hit, attrib_keys, min_pos, genom_loc, ovl_pf, ovl_fp, i):
    """Saves info of the hit in a tabular form to be written in the output table. """
    if attrib_keys != ["None"]:
        # extract min_dist_content
        [dist, [feat, fstart, fend, strand, attrib_val]] = min_dist_hit
        # attrib_val.strip("\r").strip("\t").strip("\n")
        dist = max(dist) if isinstance(dist, list) else dist
        dist = '%d' % round(dist, 1)
        best_res = "\t".join(np.hstack([peak_id, chrom, pstart, p_center, pend, feat, fstart,
                                        fend, strand, min_pos, dist, genom_loc, str(ovl_pf), str(ovl_fp), attrib_val, str(i)]))
        return best_res + "\n"

    elif attrib_keys == ["None"]:
        [dist, [feat, fstart, fend, strand]] = min_dist_hit
        dist = max(dist) if isinstance(dist, list) else dist
        dist = '%d' % round(dist, 1)
        best_res = "\t".join([peak_id, chrom, pstart, p_center, pend, feat, fstart,
                              fend, strand, min_pos, dist, genom_loc, str(ovl_pf), str(ovl_fp), str(i)])
        return best_res + "\n"


def write_hit_to_All(All_hits_tab, p_name, attrib_k, Dhit, hit, peak_id, chrom, pstart, pend, p_center, min_pos, genomic_location, ovl_pk, ovl_feat, j):
    """Writes an output table."""
    if attrib_k != ["None"]:
        attr_val = [get_hit_attribute(hit, a) for a in attrib_k]
        hit2add = [Dhit, [hit[2], hit[3], hit[4], hit[6], attr_val]]
        the_res = create_table(peak_id, chrom, pstart, pend, p_center,
                               hit2add, attrib_k, min_pos, genomic_location, ovl_pk, ovl_feat, j)
    elif attrib_k == ["None"]:
        hit2add = [Dhit, [hit[2], hit[3], hit[4], hit[6]]]
        the_res = create_table(peak_id, chrom, pstart, pend, p_center,
                               hit2add, attrib_k, min_pos, genomic_location, ovl_pk, ovl_feat, j)

    if All_hits_tab[j][p_name] == '' or All_hits_tab[j][p_name].split("\t")[9] == "NA":
        All_hits_tab[j][p_name] = the_res

    elif All_hits_tab[j][p_name] != '' or All_hits_tab[j][p_name].split("\t")[9] != "NA":
        All_hits_tab[j][p_name] = All_hits_tab[j][p_name].strip("\n")
        All_hits_tab[j][p_name] = '\n'.join((All_hits_tab[j][p_name], the_res))

    return All_hits_tab[j][p_name]


def concat_comments(queries, priority, annot_gtf, peaks_bed):
    """Puts together comments for file header."""
    comments = ["#UROPA-Universal RObust Peak Annotator"]
    for q in enumerate(queries):
        comments.append("#Query No {} :".format(q[0]))
        comments.append("#feature: {}, strand: {}, feature.anchor: {}, distance : {} , direction : {}, internals :{},\n#show.attribute(s): {}, filter.attribute: {},attribute.value:{}".format(
                q[1]['feature'], q[1]['strand'], q[1]['feature.anchor'], q[
                    1]['distance'], q[1]['direction'], q[1]['internals'],
                q[1]['show.attributes'], q[1]['filter.attribute'], q[1]['attribute.value']))
    comments.append("#priority:{}".format(priority))
    comments.append("#GTF: {}".format(annot_gtf))
    comments.append("#BED: {}".format(peaks_bed))
    comments.append("#Columns.terminology:")
    comments.append("#feature, feature_start, feature_end, feature_strand: The information of the genomic feature that annotates the peak, as extracted by the gtf file.")
    comments.append(
            "#distance:The distance measured as following: abs(peak.center-feature.anchor).If no feature.anchor given,then the minimum of 3 distances from each feature.anchor{start,center,end} to peak.center is chosen.")
    comments.append("#feat_anchor : The position of the genomic feature chosen for annotation that had the minimum distance to the peak.center.If feature.anchor given in config this will be shown also here.")
    comments.append(
            "#genomic_location: The position of the peak relative to the annotated feature direction.")
    comments.append("#feat_ovl_peak : When peak and feature overlap(i.e genomic_location = overlapStart), Ratio(overlapping region / peak.length) shows percentage of peak covered by the feature.(i.e 1.0 = 100% of peak covered, peak is internal.")
    comments.append("#peak_ovl_feat : When peak and feature overlap(i.e genomic_location = overlapStart), Ratio(overlapping region / feature.length) shows percentage of feature covered by the peak.(i.e 1.0 = 100% of feature covered, feature is internal.)")
    comments.append("#Attributes that have been given in the key 'show.atttributes' will be shown here and their values extracted by the gtf will be displayed for each feature.If 'filter.attribute' contains same attribute, this column helps confirm the filtering.")
    comments.append("#query:The query that validates with its given parameters the feature to be assigned to the peak.If only one query given, column will always display '0',the first query.")
    comments.append("#---------------------------------------------------------------------------------------------------------------------------------------#")

    return '\n'.join(comments)

def write_partial_file(outfile, Table):
    """Writes a partial file."""
    with open(outfile, "w") as of:
        for k in Table:
            of.write("".join(Table[k]))


def merge_queries(Best_combo_k):
    """Merges queries."""
    merged_q = [Best_combo_k[l].split(
        "\t")[-1].strip("\n") for l in range(len(Best_combo_k))]
    # Read if combo_k is "" and give it the q.numb, so I have it ready for
    # merging when one query =""
    merged_q[-1] = merged_q[-1] + "\n"
    joined_q = ",".join(merged_q)
    Best_combo_k_0 = "\t".join(Best_combo_k[0].split("\t")[:-1])
    Best_merg_q = Best_combo_k_0 + "\t" + joined_q
    return Best_merg_q

def finalize_file(ffile, partials, header, comments=None, log=None):
    """Writes a output table of UROPA."""
    try:
        with open(ffile, "w") as f:
            if not comments is None:
                f.write(comments+'\n')
            f.write(header+'\n')
            for p in partials:
                with open(p) as pf:
                    shutil.copyfileobj(pf, f)
    except IOError:
        if not log is None:
            log.error("Unable to write to file {}".format(ffile))

def cleanup(directory, log=None):
    """Removes temporary files from a UROPA directory."""
    temp_files = glob.glob(directory + '*_part_*')
    for f in temp_files:
        try:
            os.remove(f)
        except OSError:
            if not log is None:
                log.error("Could not remove temporary file {}".format(f))
            continue



from setuptools import setup

def readme():
    with open('README.md', encoding='utf-8') as f:
        return f.read()

setup(name='uropa',
      version='2.0.0-alpha',
      description='UROPA is a command line based tool, intended for genomic region annotation',
      long_description=readme(),
      url='https://github.molgen.mpg.de/loosolab/UROPA',
      author='Jens Preussner',
      author_email='jens.preussner@mpi-bn.mpg.de',
      license='MIT',
      packages=['uropa'],
      entry_points = {
        'console_scripts': ['uropa = uropa.uropa:main']
      },
      scripts = ['utils/uropa_summary.R', 'utils/uropa_reformat_output.R', 'utils/uropa2gtf.R'],
      install_requires=[
        'numpy',
        'pysam'
      ],
      classifiers = [
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Programming Language :: Python :: 3.4'
      ],
      zip_safe=False,
      include_package_data=True)
